﻿using HeuristicLab.Operators;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Analysis;

namespace HeuristicLab.Algorithms.SimulatedAnnealing
{
    [Item("ConnollyReheatingOperator", "Reheats the temperature only once after N consecutive solutions are rejected. It reheats to the temperature where the best solutions was found and the temperature stays there.")]
    [StorableClass]
    public class ConnollyReheatingOperator : SingleSuccessorOperator, IReheatingOperator
    {
        #region Strings
        private const string AnnealingOperatorName = "AnnealingOperator";
        private const string LowerTemperatureName = "LowerTemperature";
        private const string IterationsName = "Iterations";
        private const string TemperatureStartIndexName = "TemperatureStartIndex";
        private const string StartTemperatureName = "StartTemperature";
        private const string EndTemperatureName = "EndTemperature";
        private const string TemperatureName = "Temperature";
        private const string MaximumIterationsName = "MaximumIterations";

        private const string ThresholdName = "Threshold";
        private const string IdealTemperatureName = "IdealTemperature";
        private const string AcceptanceMemoryName = "AcceptanceMemory";
        private const string IsAcceptedName = "IsAccepted";
        private const string QualityName = "Quality";
        private const string CoolingName = "Cooling";
        private const string ResultsName = "Results";

        private const string BestQualityName = "BestQuality";
        private const string TemperatureChartName = "Temperature Chart";
        private const string QualitiesName = "Qualities";
        private const string OptimalTemperatureName = "Optimal Temperature";

        #endregion
        #region Parameters
        public ILookupParameter<DoubleValue> TemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[TemperatureName]; }
        }
        public ILookupParameter<DoubleValue> LowerTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[LowerTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> StartTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[StartTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> EndTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[EndTemperatureName]; }
        }
        public ILookupParameter<IntValue> TemperatureStartIndexParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[TemperatureStartIndexName]; }
        }
        public ILookupParameter<IntValue> IterationsParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[IterationsName]; }
        }
        public ILookupParameter<IDiscreteDoubleValueModifier> AnnealingOperatorParameter
        {
            get { return (ILookupParameter<IDiscreteDoubleValueModifier>)Parameters[AnnealingOperatorName]; }
        }
        public ILookupParameter<IntValue> MaximumIterationsParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[MaximumIterationsName]; }
        }
        private ValueParameter<IntValue> ThresholdParameter
        {
            get { return (ValueParameter<IntValue>)Parameters[ThresholdName]; }
        }
        public ILookupParameter<BoolValue> IsAcceptedParameter
        {
            get { return (ILookupParameter<BoolValue>)Parameters[IsAcceptedName]; }
        }
        public ILookupParameter<ItemList<BoolValue>> AcceptanceMemoryParameter
        {
            get { return (ILookupParameter<ItemList<BoolValue>>)Parameters[AcceptanceMemoryName]; }
        }
        public ILookupParameter<BoolValue> CoolingParameter
        {
            get { return (ILookupParameter<BoolValue>)Parameters[CoolingName]; }
        }

        #endregion

        public ConnollyReheatingOperator() : base()
        {
            #region Create parameters
            Parameters.Add(new LookupParameter<DoubleValue>(TemperatureName, "The current temperature."));
            Parameters.Add(new LookupParameter<DoubleValue>(LowerTemperatureName, "The lower bound of the temperature."));
            Parameters.Add(new LookupParameter<IntValue>(IterationsName, "The number of iterations."));
            Parameters.Add(new LookupParameter<IDiscreteDoubleValueModifier>(AnnealingOperatorName, "The operator that cools the temperature."));
            Parameters.Add(new LookupParameter<IntValue>(TemperatureStartIndexName, "The index where the annealing or heating was last changed."));
            Parameters.Add(new LookupParameter<DoubleValue>(StartTemperatureName, "The temperature from which cooling or reheating should occur."));
            Parameters.Add(new LookupParameter<DoubleValue>(EndTemperatureName, "The temperature to which should be cooled or heated."));
            Parameters.Add(new LookupParameter<IntValue>(MaximumIterationsName, "The maximum number of iterations which should be processed."));
            Parameters.Add(new LookupParameter<BoolValue>(IsAcceptedName, "Whether the move was accepted or not."));
            Parameters.Add(new LookupParameter<ItemList<BoolValue>>(AcceptanceMemoryName, "Memorizes the last N acceptance decisions."));
            Parameters.Add(new ValueParameter<IntValue>(ThresholdName, "How many consecutive solutions have to be rejected to trigger the reheat.", new IntValue(10)));
            Parameters.Add(new LookupParameter<BoolValue>(CoolingName, "True when the temperature should be cooled, false otherwise."));
            #endregion

            Parameterize();
        }

        [StorableConstructor]
        protected ConnollyReheatingOperator(bool deserializing) : base(deserializing) { }
        protected ConnollyReheatingOperator(ConnollyReheatingOperator original, Cloner cloner) : base(original, cloner) { }


        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new ConnollyReheatingOperator(this, cloner);
        }

        public void Parameterize()
        {
        }

        public override IOperation Apply()
        {
            var cooling = CoolingParameter.ActualValue.Value;

            if(cooling)
            {
                var acceptances = AcceptanceMemoryParameter.ActualValue;
                acceptances.Add(IsAcceptedParameter.ActualValue);

                if (acceptances.Count > ThresholdParameter.Value.Value) acceptances.RemoveAt(0);
                var stopCooling = acceptances.Count == ThresholdParameter.Value.Value && !acceptances.Any(x => x.Value);

                if(stopCooling)
                {
                    CoolingParameter.ActualValue.Value = false;
                    return ReheatToIdealTemperature();
                }else
                {
                    return Cool();
                }
            }
            else
            {
                return KeepTemperature();
            }
        }

        private IOperation ReheatToIdealTemperature()
        {
            var algorithm = ExecutionContext.Parent.Parent.Parent;
            var results = (ResultCollection)algorithm.Scope.Variables[ResultsName].Value;
            var bestQualitySoFar = (DoubleValue)results[BestQualityName].Value;
            var temperatures = (DataTable)results[TemperatureChartName].Value;
            var qualities = (DataTable)results[QualitiesName].Value;
            var bestQualities = qualities.Rows[BestQualityName];
            var indexOfBestQuality = bestQualities.Values.IndexOf(bestQualitySoFar.Value);
            var bestTemperature = temperatures.Rows[TemperatureName].Values[indexOfBestQuality];

            TemperatureParameter.ActualValue.Value = bestTemperature;

            if (!results.ContainsKey(OptimalTemperatureName))
            {
                results.Add(new Result(OptimalTemperatureName, new DoubleValue(bestTemperature)));
            }
            else
            {
                results[OptimalTemperatureName].Value = new DoubleValue(bestTemperature);
            }
            return base.Apply();
        }

        private IOperation Cool()
        {
            return new OperationCollection {
                    ExecutionContext.CreateOperation(AnnealingOperatorParameter.ActualValue),
                    base.Apply()
                };
        }

        private IOperation KeepTemperature()
        {
            return base.Apply();
        }
    }
}
